import random
import math
import numpy as np
import pytweening


class HumanizeMouseTrajectory:
    def __init__(self, from_point, to_point, **kwargs):
        self.from_point = from_point
        self.to_point = to_point
        self.points = self.generate_curve(**kwargs)

    def generate_curve(self, **kwargs):
        """Generates the curve based on arguments below, default values below are automatically modified to cause randomness"""
        offset_boundary_x = kwargs.get("offset_boundary_x", 80)
        offset_boundary_y = kwargs.get("offset_boundary_y", 80)
        left_boundary = (
            kwargs.get("left_boundary", min(self.from_point[0], self.to_point[0]))
            - offset_boundary_x
        )
        right_boundary = (
            kwargs.get("right_boundary", max(self.from_point[0], self.to_point[0]))
            + offset_boundary_x
        )
        down_boundary = (
            kwargs.get("down_boundary", min(self.from_point[1], self.to_point[1]))
            - offset_boundary_y
        )
        up_boundary = (
            kwargs.get("up_boundary", max(self.from_point[1], self.to_point[1]))
            + offset_boundary_y
        )
        knots_count = kwargs.get("knots_count", 2)
        distortion_mean = kwargs.get("distortion_mean", 1)
        distortion_st_dev = kwargs.get("distortion_st_dev", 1)
        distortion_frequency = kwargs.get("distortion_frequency", 0.5)
        tween = kwargs.get("tweening", pytweening.easeOutQuad)
        target_points = kwargs.get("target_points", 100)

        internalKnots = self.generate_internal_knots(
            left_boundary, right_boundary, down_boundary, up_boundary, knots_count
        )
        points = self.generate_points(internalKnots)
        points = self.distort_points(
            points, distortion_mean, distortion_st_dev, distortion_frequency
        )
        points = self.tween_points(points, tween, target_points)
        return points

    def generate_internal_knots(
        self, l_boundary, r_boundary, d_boundary, u_boundary, knots_count
    ):
        """Generates the internal knots of the curve randomly"""
        if not (
            self.check_if_numeric(l_boundary)
            and self.check_if_numeric(r_boundary)
            and self.check_if_numeric(d_boundary)
            and self.check_if_numeric(u_boundary)
        ):
            raise ValueError("Boundaries must be numeric values")
        if not isinstance(knots_count, int) or knots_count < 0:
            knots_count = 0
        if l_boundary > r_boundary:
            raise ValueError(
                "left_boundary must be less than or equal to right_boundary"
            )
        if d_boundary > u_boundary:
            raise ValueError(
                "down_boundary must be less than or equal to upper_boundary"
            )
        try:
            knotsX = np.random.choice(range(l_boundary, r_boundary) or l_boundary, size=knots_count)
            knotsY = np.random.choice(range(d_boundary, u_boundary) or d_boundary, size=knots_count)
        except TypeError:
            knotsX = np.random.choice(
                range(int(l_boundary), int(r_boundary)), size=knots_count
            )
            knotsY = np.random.choice(
                range(int(d_boundary), int(u_boundary)), size=knots_count
            )
        knots = list(zip(knotsX, knotsY))
        return knots

    def generate_points(self, knots):
        """Generates the points from BezierCalculator"""
        if not self.check_if_list_of_points(knots):
            raise ValueError("knots must be valid list of points")

        midPtsCnt = max(
            abs(self.from_point[0] - self.to_point[0]),
            abs(self.from_point[1] - self.to_point[1]),
            2,
        )
        knots = [self.from_point] + knots + [self.to_point]
        return BezierCalculator.calculate_points_in_curve(int(midPtsCnt), knots)

    def distort_points(
        self, points, distortion_mean, distortion_st_dev, distortion_frequency
    ):
        """Distorts points by parameters of mean, standard deviation and frequency"""
        if not (
            self.check_if_numeric(distortion_mean)
            and self.check_if_numeric(distortion_st_dev)
            and self.check_if_numeric(distortion_frequency)
        ):
            raise ValueError("Distortions must be numeric")
        if not self.check_if_list_of_points(points):
            raise ValueError("points must be valid list of points")
        if not (0 <= distortion_frequency <= 1):
            raise ValueError("distortion_frequency must be in range [0,1]")

        distorted = []
        for i in range(1, len(points) - 1):
            x, y = points[i]
            delta = (
                np.random.normal(distortion_mean, distortion_st_dev)
                if random.random() < distortion_frequency
                else 0
            )
            # distorted += ((x, y + delta),)
            distorted.append((x, y + delta))
        distorted = [points[0]] + distorted + [points[-1]]
        return distorted

    def tween_points(self, points, tween, target_points):
        """Modifies points by tween"""
        if not self.check_if_list_of_points(points):
            raise ValueError("List of points not valid")
        if not isinstance(target_points, int) or target_points < 2:
            raise ValueError("target_points must be an integer greater or equal to 2")

        res = []
        for i in range(target_points):
            index = int(tween(float(i) / (target_points - 1)) * (len(points) - 1))
            res.append(points[index])
        return res

    @staticmethod
    def check_if_numeric(val):
        """Checks if value is proper numeric value"""
        return isinstance(val, (float, int, np.int32, np.int64, np.float32, np.float64))

    def check_if_list_of_points(self, list_of_points):
        """Checks if list of points is valid"""
        if not isinstance(list_of_points, list):
            return False
        try:
            point = lambda p: (
                (len(p) == 2)
                and self.check_if_numeric(p[0])
                and self.check_if_numeric(p[1])
            )
            return all(map(point, list_of_points))
        except (KeyError, TypeError):
            return False


class BezierCalculator:
    @staticmethod
    def binomial(n, k):
        """Returns the binomial coefficient "n choose k" """
        return math.factorial(n) / float(math.factorial(k) * math.factorial(n - k))

    @staticmethod
    def bernstein_polynomial_point(x, i, n):
        """Calculate the i-th component of a bernstein polynomial of degree n"""
        return BezierCalculator.binomial(n, i) * (x**i) * ((1 - x) ** (n - i))

    @staticmethod
    def bernstein_polynomial(points):
        """
        Given list of control points, returns a function, which given a point [0,1] returns
        a point in the Bézier curve described by these points
        """

        def bernstein(t):
            n = len(points) - 1
            x = y = 0
            for i, point in enumerate(points):
                bern = BezierCalculator.bernstein_polynomial_point(t, i, n)
                x += point[0] * bern
                y += point[1] * bern
            return x, y

        return bernstein

    @staticmethod
    def calculate_points_in_curve(n, points):
        """
        Given list of control points, returns n points in the Bézier curve,
        described by these points
        """
        curvePoints = []
        bernstein_polynomial = BezierCalculator.bernstein_polynomial(points)
        for i in range(n):
            t = i / (n - 1)
            curvePoints.append((bernstein_polynomial(t)))
        return curvePoints


import random

from botasaurus_driver import Driver, cdp
from .human_curve_generator import HumanizeMouseTrajectory
from .calculate_and_randomize import generate_random_curve_parameters, calculate_absolute_offset


class WebAdjuster:
    def __init__(self, driver: Driver):
        """
        Initialize WebAdjuster with a Botasaurus driver
        
        Args:
            driver: Botasaurus Driver instance
        """
        self.driver = driver
        self.origin_coordinate = [0, 0]

    def do_move(self, x,y):
        # print(x,y)
        self.driver.run_cdp_command(cdp.input_.dispatch_mouse_event(
                            "mouseMoved",
                            x=x,
                            y=y
                    ))        
    def move_to(
        self,
        element_or_pos,
        origin_coordinates=None,
        absolute_offset=False,
        relative_position=None,
        human_curve=None,
        steady=False
    ):
        """Moves the cursor, trying to mimic human behaviour!"""
        origin = origin_coordinates
        if origin_coordinates is None:
            origin = self.origin_coordinate

        pre_origin = tuple(origin)
        if isinstance(element_or_pos, (list,tuple)):
            if not absolute_offset:
                x, y = element_or_pos[0], element_or_pos[1]
            else:
                x, y = (
                    element_or_pos[0] + pre_origin[0],
                    element_or_pos[1] + pre_origin[1],
                )
        else:
            # Get element position using Botasaurus's get_bounding_rect, with fallback if needed
            try:
                rect = element_or_pos.get_bounding_rect()
            except Exception as e:
                print("Error obtaining bounding rect for element:", e)
                if hasattr(element_or_pos, "_elem"):
                    raw = element_or_pos._elem
                    rect = self.driver.run_js("(function(el){var r=el.getBoundingClientRect(); return {x: r.x, y: r.y, width: r.width, height: r.height};})(arguments[0]);", [raw])
                else:
                    print("Element does not support _elem attribute, cannot get position.")
                    return origin
            if rect.get("width", 0) == 0 or rect.get("height", 0) == 0:
                print("Could not find position for", element_or_pos)
                return origin
            destination = {"x": rect.get("x", 0), "y": rect.get("y", 0)}
            
            if relative_position is None:
                x_random_off = random.choice(range(20, 80)) / 100
                y_random_off = random.choice(range(20, 80)) / 100

                # Get element size from bounding rect
                element_width = rect["width"]
                element_height = rect["height"]
                
                x, y = destination["x"] + (
                    element_width * x_random_off
                ), destination["y"] + (element_height * y_random_off)
            else:
                abs_exact_offset = calculate_absolute_offset(
                    element_or_pos, relative_position, rect
                )
                x_exact_off, y_exact_off = abs_exact_offset[0], abs_exact_offset[1]
                x, y = destination["x"] + x_exact_off, destination["y"] + y_exact_off

        (
            offset_boundary_x,
            offset_boundary_y,
            knots_count,
            distortion_mean,
            distortion_st_dev,
            distortion_frequency,
            tween,
            target_points,
        ) = generate_random_curve_parameters(
            self.driver, [origin[0], origin[1]], [x, y]
        )
        if steady:
            offset_boundary_x, offset_boundary_y = 10, 10
            distortion_mean, distortion_st_dev, distortion_frequency = 1.2, 1.2, 1
        if not human_curve:
            human_curve = HumanizeMouseTrajectory(
                [origin[0], origin[1]],
                [x, y],
                offset_boundary_x=offset_boundary_x,
                offset_boundary_y=offset_boundary_y,
                knots_count=knots_count,
                distortion_mean=distortion_mean,
                distortion_st_dev=distortion_st_dev,
                distortion_frequency=distortion_frequency,
                tween=tween,
                target_points=target_points,
            )
        for point in human_curve.points:
            # Move to each point in the curve
            self.do_move(
                    x=point[0],
                    y=point[1]
            )
            
            # Update the origin coordinates
            origin[0], origin[1] = point[0], point[1]
        self.origin_coordinate = [x, y]
        return [x, y]


from time import sleep
import random
from typing import Union
from botasaurus_driver import Driver, cdp, BrowserTab
from botasaurus_driver.driver import Element
from .web_adjuster import WebAdjuster

class WebCursor:
    def __init__(self, driver: Driver):
        """
        Initialize WebCursor with a Botasaurus driver
        
        Args:
            driver: Botasaurus Driver instance
        """
        self.driver = driver
        self.human = WebAdjuster(self.driver)
        self.origin_coordinates = [0, 0]

    def move_to(
            self,
            element: Union[Element, list],
            relative_position: list = None,
            absolute_offset: bool = False,
            origin_coordinates=None,
            steady=False
    ):
        """Moves to element or coordinates with human curve"""
        if not self.scroll_into_view_of_element(element):
            return False
        if origin_coordinates is None:
            origin_coordinates = self.origin_coordinates
        self.origin_coordinates = self.human.move_to(
            element,
            origin_coordinates=origin_coordinates,
            absolute_offset=absolute_offset,
            relative_position=relative_position,
            steady=steady
        )
        return self.origin_coordinates

    def click(
            self,
            element: Union[Element, list],
            number_of_clicks: int = 1,
            click_duration: float = 0,
            relative_position: list = None,
            absolute_offset: bool = False,
            origin_coordinates=None,
            steady=False
    ):
        """Moves to element or coordinates with human curve, and clicks on it a specified number of times, default is 1"""
        self.move_to(
            element,
            origin_coordinates=origin_coordinates,
            absolute_offset=absolute_offset,
            relative_position=relative_position,
            steady=steady
        )
        self.random_natural_sleep()
        self._click(number_of_clicks=number_of_clicks, click_duration=click_duration)
        return True

    def random_natural_sleep(self):
        sleep(random.randint(170, 280) / 1000)

    def _click(self, number_of_clicks: int = 1, click_duration: float = 0):
        """Performs the click action"""
        for _ in range(number_of_clicks):
                # Using CDP to perform mousedown
                self.driver.run_cdp_command(
                    cdp.input_.dispatch_mouse_event(
                        "mousePressed",
                        x=self.origin_coordinates[0],
                        y=self.origin_coordinates[1],
                        button=cdp.input_.MouseButton("left"),
                        click_count=1,
                    )
                )
                if click_duration:
                    sleep(click_duration)
                else:
                    self.random_natural_sleep()
                # Using CDP to perform mouseup and click
                self.driver.run_cdp_command(
                    cdp.input_.dispatch_mouse_event(
                        "mouseReleased",
                        x=self.origin_coordinates[0],
                        y=self.origin_coordinates[1],
                        button=cdp.input_.MouseButton("left"),
                        click_count=1,
                        )
                )
                self.random_natural_sleep()
        return True

    def move_by_offset(self, x: int, y: int, steady=False):
        """Moves the cursor with human curve, by specified number of x and y pixels"""
        self.origin_coordinates = self.human.move_to([x, y],  absolute_offset=True, steady=steady)
        return True

    def drag_and_drop(
            self,
            drag_from_element: Union[Element, list],
            drag_to_element: Union[Element, list],
            drag_from_relative_position: list = None,
            drag_to_relative_position: list = None,
            steady=False
    ):
        """Moves to element or coordinates, clicks and holds, dragging it to another element, with human curve"""
        if drag_from_relative_position is None:
            self.move_to(drag_from_element, steady=steady)
        else:
            self.move_to(
                drag_from_element, relative_position=drag_from_relative_position, steady=steady
            )

        if drag_to_element is None:
            self._click()
        else:
        # Click and hold using CDP
            self.driver.run_cdp_command(
                cdp.input_.dispatch_mouse_event(
                    "mousePressed",
                    x=self.origin_coordinates[0],
                    y=self.origin_coordinates[1],
                    button=cdp.input_.MouseButton("left"),
                    click_count=1
                )
            )
            if drag_to_relative_position is None:
                self.move_to(drag_to_element, steady=steady)
            else:
                self.move_to(
                    drag_to_element, relative_position=drag_to_relative_position, steady=steady
                )
            self.driver.run_cdp_command(
                cdp.input_.dispatch_mouse_event(
                    "mouseReleased",
                    x=self.origin_coordinates[0],
                    y=self.origin_coordinates[1],
                        button=cdp.input_.MouseButton("left"),
                        click_count=1,
                )
            )
        return True

    def control_scroll_bar(
            self,
            scroll_bar_element: Union[Element, list],
            amount_by_percentage: list,
            orientation: str = "horizontal",
            steady=False
    ):
        """Adjusts any scroll bar on the webpage, by the amount you want in float number from 0 to 1
        representing percentage of fullness, orientation of the scroll bar must also be defined by user
        horizontal or vertical"""
        direction = True if orientation == "horizontal" else False

        self.move_to(scroll_bar_element)
        # Click and hold using CDP
        self.driver.run_cdp_command(
            cdp.input_.dispatch_mouse_event(
                "mousePressed",
                x=self.origin_coordinates[0],
                y=self.origin_coordinates[1],
                        button=cdp.input_.MouseButton("left"),
                        click_count=1,
            )
        )
        
        # Move to target position
        if direction:
            self.move_to(
                scroll_bar_element,
                relative_position=[amount_by_percentage, random.randint(0, 100) / 100],
                steady=steady
            )
        else:
            self.move_to(
                scroll_bar_element,
                relative_position=[random.randint(0, 100) / 100, amount_by_percentage],
                steady=steady
            )

        # Release using CDP
        self.driver.run_cdp_command(
            cdp.input_.dispatch_mouse_event(
                "mouseReleased",
                x=self.origin_coordinates[0],
                y=self.origin_coordinates[1],
                        button=cdp.input_.MouseButton("left"),
                        click_count=1,
            )
        )

        return True

    def scroll_into_view_of_element(self, element: Element):
        """Scrolls the element into viewport, if not already in it"""
        if isinstance(element, Element):
            self.scroll_to_element(element)
            return True
        elif isinstance(element, (list, tuple)):
            element = self.driver.get_element_at_point(x=element[0],y=element[1])
            # an iframe, no need to scroll into view
            if isinstance(element, BrowserTab):
              return True

            self.scroll_to_element(element)
            return True
        else:
            print("Incorrect Element or Coordinates values!")
            return False

    def scroll_to_element(self, element):
        scrolling_code = """(element)=>{
              var rect = element.getBoundingClientRect();
              return (
                rect.top >= 0 &&
                rect.left >= 0 &&
                rect.bottom <= (window.innerHeight || document.documentElement.clientHeight) &&
                rect.right <= (window.innerWidth || document.documentElement.clientWidth)
              );            
}"""
        is_in_viewport = element.run_js(scrolling_code)
        if not is_in_viewport:
            element.scroll_into_view()
            sleep(random.uniform(0.8, 1.4))

    def show_cursor(self):
        self.driver.run_js('''if (!window.dEl) {
    let dot
    function displayRedDot() {
        // Get the cursor position
        const x = event.clientX
        const y = event.clientY

        if (!dot) {
            // Create a new span element for the red dot if it doesn't exist
            dot = document.createElement("span")
            // Style the dot with CSS
            dot.style.position = "fixed"
            dot.style.width = "5px"
            dot.style.height = "5px"
            dot.style.borderRadius = "50%"
            dot.style.backgroundColor = "#E7010A"
            // Add the dot to the page
            document.body.appendChild(dot)
        }

        // Update the dot's position
        dot.style.left = x + "px"
        dot.style.top = y + "px"
    }

    // Add event listener to update the dot's position on mousemove
    document.addEventListener("mousemove", displayRedDot)
    window.dEl = dot
}''')

    def hide_cursor(self):
        self.driver.run_js('''window.dEl?.remove()''')


import random
import pytweening


def calculate_absolute_offset(element, relative_position, rect=None):
    """Calculates the absolute offset based on relative position"""
    if rect is None:
        # Get element position using get_bounding_rect
        rect = element.get_bounding_rect()
    
    element_width = rect["width"]
    element_height = rect["height"]
    
    x_exact_off = element_width * relative_position[0]
    y_exact_off = element_height * relative_position[1]
    
    return [int(x_exact_off), int(y_exact_off)] 


def generate_random_curve_parameters(driver, pre_origin, post_destination):
    """Generates random parameters for the curve, the tween, number of knots, distortion, target points and boundaries"""
    
    # Get window size using Botasaurus's run_js
    window_size = driver.run_js("""
        return {
            width: document.documentElement.clientWidth || window.innerWidth || document.body.clientWidth,
            height: document.documentElement.clientHeight || window.innerHeight || document.body.clientHeight
        };
    """)

    web = True
    viewport_width, viewport_height = window_size['width'],window_size['height']
    
    min_width, max_width = viewport_width * 0.15, viewport_width * 0.85
    min_height, max_height = viewport_height * 0.15, viewport_height * 0.85

    tween_options = [
        pytweening.easeOutExpo,
        pytweening.easeInOutQuint,
        pytweening.easeInOutSine,
        pytweening.easeInOutQuart,
        pytweening.easeInOutExpo,
        pytweening.easeInOutCubic,
        pytweening.easeInOutCirc,
        pytweening.linear,
        pytweening.easeOutSine,
        pytweening.easeOutQuart,
        pytweening.easeOutQuint,
        pytweening.easeOutCubic,
        pytweening.easeOutCirc,
    ]

    tween = random.choice(tween_options)
    offset_boundary_x = random.choice(
        random.choices(
            [range(20, 45), range(45, 75), range(75, 100)], [0.2, 0.65, 15]
        )[0]
    )
    offset_boundary_y = random.choice(
        random.choices(
            [range(20, 45), range(45, 75), range(75, 100)], [0.2, 0.65, 15]
        )[0]
    )
    knots_count = random.choices(
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        [0.15, 0.36, 0.17, 0.12, 0.08, 0.04, 0.03, 0.02, 0.015, 0.005],
    )[0]

    distortion_mean = random.choice(range(80, 110)) / 100
    distortion_st_dev = random.choice(range(85, 110)) / 100
    distortion_frequency = random.choice(range(25, 70)) / 100

    target_points = random.choice(
            random.choices(
                [range(35, 45), range(45, 60), range(60, 80)], [0.53, 0.32, 0.15]
            )[0]
        )
    

    if (
            min_width > pre_origin[0]
            or max_width < pre_origin[0]
            or min_height > pre_origin[1]
            or max_height < pre_origin[1]
    ):
        offset_boundary_x = 1
        offset_boundary_y = 1
        knots_count = 1
    if (
            min_width > post_destination[0]
            or max_width < post_destination[0]
            or min_height > post_destination[1]
            or max_height < post_destination[1]
    ):
        offset_boundary_x = 1
        offset_boundary_y = 1
        knots_count = 1
    return (
        offset_boundary_x,
        offset_boundary_y,
        knots_count,
        distortion_mean,
        distortion_st_dev,
        distortion_frequency,
        tween,
        target_points,
    )
